from typing import List
import importlib
import os
import re
from traceback import format_exception
from sys import exc_info

class PluginManager():
  @classmethod
  def get_plugins(cls, chain:str, path:str) -> List:
    try:
      plugins = []
      files = os.listdir(path)
      dirs = sorted([f for f in files if os.path.isdir(os.path.join(path, f)) and "_plugin" in f and "." not in f])
      for dir in dirs:
        module = importlib.import_module(f"plugin.{dir}.{dir}", "plugin")
        dir = list(dir)
        dir[0] = dir[0].upper()
        dir = "".join(dir)
        dir = re.sub("_(.)",lambda x:x.group(1).upper(), dir)
        plugin_class = getattr(module, dir)
        if plugin_class.chain.lower() == chain.lower(): plugins.append(plugin_class)
      return plugins
    except Exception as e:
      etype, value, tb= exc_info()
      info= format_exception(etype, value, tb)[-2]
      raise RuntimeError(f"Failed to get plugins. check {path} has appropriate senka plugins. Exception in{info}")


from decimal import Decimal
from senkalib.caaj_plugin import CaajPlugin
from senkalib.caaj_journal import CaajJournal


PANCAKESWAP_ADDRESS_TRADE = "0x10ED43C718714eb63d5aA57B78B54704E256024E"
PANCAKESWAP_ADDRESS_EARN = "0x73feaa1eE314F8c655E354234017bE2193C9E24E"
PANCAKESWAP_SYRUP_ADDRESS = "0x009cf7bc57584b7998236eff51b98a168dcea9b0"

CAKE_CONTRACT_ADDRESS = '0x0E09FaBB73Bd3Ade0a17ECC321fD13a19e81cE82'
WBNB_CONTRACT_ADDRESS = '0xbb4cdb9cbd36b01bd1cbaebf2de08d9173bc095c'
SYRUP_CONTRACT_ADDRESS = '0x009cF7bC57584b7998236eff51b98A168DceA9B0'

PANCAKESWAP_ADDRESS = [PANCAKESWAP_ADDRESS_EARN,
                       PANCAKESWAP_ADDRESS_TRADE,
                       CAKE_CONTRACT_ADDRESS]

LP_DEPOSIT_TOPIC = "0x90890809c654f11d6e72a28fa60149770a0d11ec6c92319d6ceb2bb0a4ea1a15"
LP_WITHDROW_TOPIC = "0xf279e6a1f5e320cca91135676d9cb6e44ca8a08c0b88342bcdb1144f6511b568"


WETH_CONTRACT_ADDRESS = '0xc02aaa39b223fe8d0a0e5c4f27ead9083c756cc2'
WETH_DEPOSIT_TOPIC = '0xe1fffcc4923d04b559f4d29a8bfc6cda04eb5b0d3c460751c2402c5c5cc9109c'
WETH_WITHDRAWAL_TOPIC = '0x7fcf532c15f0a6db0bd6d0e038bea71d30d808c7d98cb3bf7268a95bf5081b65'

WETH_EARN_WITHDRAWAL_TOPIC = '0xf279e6a1f5e320cca91135676d9cb6e44ca8a08c0b88342bcdb1144f6511b568'


ERC20_TRANSFER_TOPIC = '0xddf252ad1be2c89b69c2b068fc378daa952ba7f163c4a11628f55a4df523b3ef'
ERC20_APPROVE_TOPIC = '0x8c5be1e5ebec7d5bd14f71427d1e84f3dd0314c0f7b2291e5b200ac8c7c3b925'
ERC20_BURN_TOPIC = "0xdccd412f0b1252819cb1fd330b93224ca42612892bb3f4f789976e6d81936496"
ERC20_MINT_TOPIC = "0x4c209b5fc8ad50758f13e2e1088ba56a560dff690a1c6fef26394f4c03821c4f"

WEI = 10**18


class PancakePlugin(CaajPlugin):
    chain = "Bsc"
    CHAIN_CONTRACT = "BNB"
    PLATFORM = "bnb_pancakeswap"

    @classmethod
    def can_handle(cls, transaction) -> bool:
        swap_type = transaction.transaction_receipt['to']
        if swap_type in PANCAKESWAP_ADDRESS:
            return True
        else:
            return False

    @classmethod
    def get_caajs(cls, transaction) -> CaajJournal:
        topics = list(
            map(lambda log: log['topics'][0].hex().lower(), transaction.transaction_receipt['logs']))
        if transaction.transaction_receipt['status'] == 1:
            if transaction.transaction_receipt['to'] == PANCAKESWAP_ADDRESS_TRADE:

                if ERC20_BURN_TOPIC in topics:
                    caaj_main = PancakePlugin.__get_caaj_liquidity_remove(
                        transaction, topics)
                    caaj_fee = PancakePlugin.__get_caaj_fee(
                        transaction, "pancakeswap transaction fee")
                    return [caaj_main, caaj_fee]
                elif ERC20_MINT_TOPIC in topics:
                    # liquidity add
                    caaj_main = PancakePlugin.__get_caaj_liquidity_add(
                        transaction, topics)
                    caaj_fee = PancakePlugin.__get_caaj_fee(
                        transaction, "pancakeswap transaction fee")
                    return [caaj_main, caaj_fee]

                else:
                    # exchange
                    caaj_main = PancakePlugin.__get_caaj_exchange(
                        transaction, topics)
                    caaj_fee = PancakePlugin.__get_caaj_fee(
                        transaction, "pancakeswap transaction fee")
                    return [caaj_main, caaj_fee]
            elif transaction.transaction_receipt['to'] == PANCAKESWAP_ADDRESS_EARN:
                if WETH_EARN_WITHDRAWAL_TOPIC in topics:
                    # unstake
                    caaj_main = PancakePlugin.__get_caaj_farms_unstake(
                        transaction)
                    caaj_fee = PancakePlugin.__get_caaj_fee(
                        transaction, "pancakeswap transaction fee")
                    caaj_reward = PancakePlugin.__get_caaj_farms_reward(
                        transaction)
                    return [caaj_main, caaj_fee, caaj_reward]
                else:
                    # stake and harvest
                    caaj_main = PancakePlugin.__get_caaj_farms_stake(
                        transaction)
                    caaj_fee = PancakePlugin.__get_caaj_fee(
                        transaction, "pancakeswap transaction fee")
                    return [caaj_main, caaj_fee]

    @classmethod
    def __get_caaj_fee(cls, transaction, comment):
        caaj_fee = {
            "time": transaction.get_timestamp(),
            "platform": cls.PLATFORM,
            "transaction_id": transaction.transaction_receipt['transactionHash'],
            "debit_title": "SPOT",
            "debit_from": "0x0000000000000000000000000000000000000000",
            "debit_to": transaction.transaction_receipt['from'],
            "debit_amount": {cls.CHAIN_CONTRACT: str(Decimal(transaction.get_transaction_fee())
                                                     / Decimal(WEI))},
            "credit_title": "SPOT",
            "credit_from": transaction.transaction_receipt['from'],
            "credit_to": "0x0000000000000000000000000000000000000000",
            "credit_amount": {cls.CHAIN_CONTRACT: str(Decimal(transaction.get_transaction_fee())
                                                      / Decimal(WEI))},
            "comment": comment
        }
        return caaj_fee

    @classmethod
    def __get_caaj_exchange(cls, transaction, topics):
        caaj_common = PancakePlugin.__get_caaj_common(transaction)
        if WETH_DEPOSIT_TOPIC in topics:
            credit_log_index = topics.index(WETH_DEPOSIT_TOPIC)
            debit_log = \
                list(filter(lambda log: log['topics'][0].hex().lower() == ERC20_TRANSFER_TOPIC
                            and "0x" + log['topics'][2].hex().lower()[26:] ==
                            transaction.transaction_receipt['from'].lower(),
                            transaction.transaction_receipt['logs']))[-1]
            caaj_main = {
                "debit_title": "SPOT",
                "debit_amount": {debit_log['address']:
                                 str(Decimal(int(debit_log['data'], 16))
                                 / Decimal(WEI))},
                "credit_title": "SPOT",
                "credit_amount": {cls.CHAIN_CONTRACT:
                                  str(Decimal(int(transaction.transaction_receipt['logs']
                                                  [credit_log_index]['data'], 16))
                                      / Decimal(WEI))
                                  },
                "comment": "pancakeswap trade"
            }
            caaj_main.update(caaj_common)
            return caaj_main
        elif WETH_WITHDRAWAL_TOPIC in topics:
            debit_log_index = topics.index(WETH_WITHDRAWAL_TOPIC)
            credit_log = \
                list(filter(lambda log: log['topics'][0].hex().lower() == ERC20_TRANSFER_TOPIC
                            and "0x" + log['topics'][1].hex().lower()[26:] ==
                            transaction.transaction_receipt['from'].lower(),
                            transaction.transaction_receipt['logs']))[0]
            caaj_main = {
                "debit_title": "SPOT",
                "debit_amount": {cls.CHAIN_CONTRACT:
                                 str(Decimal(int(transaction.transaction_receipt['logs']
                                                 [debit_log_index]['data'], 16))
                                     / Decimal(WEI))
                                 },
                "credit_title": "SPOT",
                "credit_amount": {credit_log['address']:
                                  str(Decimal(int(credit_log['data'], 16))
                                      / Decimal(WEI))},
                "comment": "pancakeswap trade"

            }
            caaj_main.update(caaj_common)
            return caaj_main
        else:
            credit_log = list(filter(lambda log: log['topics'][0].hex().lower() == ERC20_TRANSFER_TOPIC
                                    and f"0x{log['topics'][1].hex().lower()[26:]}" ==
                                    transaction.transaction_receipt['from']
                                    .lower(), transaction.transaction_receipt['logs']))[0]
            debit_log = list(filter(lambda log: log['topics'][0].hex().lower() == ERC20_TRANSFER_TOPIC
                                    and f"0x{log['topics'][2].hex().lower()[26:]}" ==
                                    transaction.transaction_receipt['from']
                                    .lower(), transaction.transaction_receipt['logs']))[-1]
            caaj_main = {
                "debit_title": "SPOT",
                "debit_amount": {debit_log['address']:
                                 str(Decimal(int(debit_log['data'], 16))
                                     / Decimal(WEI))
                                 },
                "credit_title": "SPOT",
                "credit_amount": {credit_log['address']:
                                  str(Decimal(int(credit_log['data'], 16))
                                      / Decimal(WEI))},
                "comment": "pancakeswap trade"

            }
            caaj_main.update(caaj_common)
            return caaj_main

    @classmethod
    def __get_caaj_liquidity_add(cls, transaction, topics):
        caaj_common = PancakePlugin.__get_caaj_common(transaction)

        if WETH_DEPOSIT_TOPIC in topics:
            # incluede bnb
            credit_log_index_1 = topics.index(WETH_DEPOSIT_TOPIC)
            credit_log_2 = \
                list(filter(lambda log: log['topics'][0].hex().lower() == ERC20_TRANSFER_TOPIC
                            and "0x" + log['topics'][1].hex().lower()[26:] ==
                            transaction.transaction_receipt['from'].lower(),
                            transaction.transaction_receipt['logs']))[0]
            debit_log = \
                list(filter(lambda log: log['topics'][0].hex().lower() == ERC20_TRANSFER_TOPIC
                            and "0x" + log['topics'][2].hex().lower()[26:] ==
                            transaction.transaction_receipt['from'].lower(),
                            transaction.transaction_receipt['logs']))[-1]
            caaj_main = {
                "debit_title": "LIQUIDITY",
                "debit_amount": {debit_log['address'].lower():
                                 str(Decimal(int(debit_log['data'].lower(), 16))
                                 / Decimal(WEI))},
                "credit_title": "SPOT",
                "credit_amount": {cls.CHAIN_CONTRACT:
                                  str(Decimal(int(transaction.transaction_receipt['logs']
                                                  [credit_log_index_1]['data']
                                      .lower(), 16)) / Decimal(WEI)),
                                  credit_log_2['address'].lower():
                                  str(Decimal(
                                      int(credit_log_2['data'].lower(), 16))/Decimal(WEI))
                                  },
                "comment": "pancakeswap add liquidity"

            }
            caaj_main.update(caaj_common)
            return caaj_main

        else:
            credit_log = list(filter(lambda log: log['topics'][0].hex().lower() == ERC20_TRANSFER_TOPIC
                                     and "0x" + log['topics'][1].hex().lower()[26:] ==
                                     transaction.transaction_receipt['from']
                                     .lower(), transaction.transaction_receipt['logs']))
            debit_log = \
                list(filter(lambda log: log['topics'][0].hex().lower() == ERC20_TRANSFER_TOPIC
                            and "0x" + log['topics'][2].hex().lower()[26:] ==
                            transaction.transaction_receipt['from'].lower(),
                            transaction.transaction_receipt['logs']))[-1]
            caaj_main = {
                "debit_title": "LIQUIDITY",
                "debit_amount": {debit_log['address']:
                                 str(Decimal(int(debit_log['data'], 16))
                                 / Decimal(WEI))},
                "credit_title": "SPOT",
                "credit_amount": {credit_log[0]['address']:
                                  str(Decimal(int(credit_log[0]['data'], 16))
                                      / Decimal(WEI)),
                                  credit_log[1]['address']:
                                  str(Decimal(
                                      int(credit_log[1]['data'], 16))/Decimal(WEI))
                                  },
                "comment": "pancakeswap add liquidity"

            }
            caaj_main.update(caaj_common)
            return caaj_main

    @ classmethod
    def __get_caaj_liquidity_remove(cls, transaction, topics):
        caaj_common = PancakePlugin.__get_caaj_common(transaction)
        if WETH_WITHDRAWAL_TOPIC in topics:
            debit_log_1_index = topics.index(WETH_WITHDRAWAL_TOPIC)
            debit_log_2 = \
                list(filter(lambda log: log['topics'][0].hex().lower() == ERC20_TRANSFER_TOPIC
                            and "0x" + log['topics'][2].hex().lower()[26:] ==
                            transaction.transaction_receipt['from'].lower(),
                            transaction.transaction_receipt['logs']))[-1]

            credit_log = \
                list(filter(lambda log: log['topics'][0].hex().lower() == ERC20_TRANSFER_TOPIC
                            and "0x" + log['topics'][1].hex().lower()[26:] ==
                            transaction.transaction_receipt['from'].lower(),
                            transaction.transaction_receipt['logs']))[0]
            caaj_main = {
                "debit_title": "SPOT",
                "debit_amount": {debit_log_2['address']:
                                 str(Decimal(int(debit_log_2['data'], 16))
                                 / Decimal(WEI)),
                                 cls.CHAIN_CONTRACT:
                                 str(Decimal(int(transaction.transaction_receipt['logs']
                                                 [debit_log_1_index]['data'], 16))/Decimal(WEI))},
                "credit_title": "LIQUIDITY",
                "credit_amount": {credit_log['address']:
                                  str(Decimal(int(credit_log['data'], 16))
                                      / Decimal(WEI))},
                "comment": "pancakeswap remove liquidity"

            }
            caaj_main.update(caaj_common)
            return caaj_main
        else:
            debit_log = \
                list(filter(lambda log: log['topics'][0].hex().lower() == ERC20_TRANSFER_TOPIC
                            and "0x" + log['topics'][2].hex().lower()[26:] ==
                            transaction.transaction_receipt['from'].lower(),
                            transaction.transaction_receipt['logs']))
            credit_log = \
                list(filter(lambda log: log['topics'][0].hex().lower() == ERC20_TRANSFER_TOPIC
                            and "0x" + log['topics'][1].hex().lower()[26:] ==
                            transaction.transaction_receipt['from'].lower(),
                            transaction.transaction_receipt['logs']))[0]
            caaj_main = {
                "debit_title": "SPOT",
                "debit_amount": {debit_log[0]['address']:
                                 str(Decimal(int(debit_log[0]['data'], 16))
                                 / Decimal(WEI)),
                                 debit_log[1]['address']:
                                 str(Decimal(int(debit_log[1]['data'], 16))
                                 / Decimal(WEI))},
                "credit_title": "LIQUIDITY",
                "credit_amount": {credit_log['address']:
                                  str(Decimal(int(credit_log['data'], 16))
                                      / Decimal(WEI))},
                "comment": "pancakeswap remove liquidity"

            }
            caaj_main.update(caaj_common)
            return caaj_main

    @classmethod
    def __get_caaj_common(cls, transaction):
        caaj_common = {
            "time": transaction.get_timestamp(),
            "platform": cls.PLATFORM,
            "transaction_id": transaction.transaction_receipt['transactionHash'],
            "debit_from": transaction.transaction_receipt['to'],
            "debit_to": transaction.transaction_receipt['from'],
            "credit_from": transaction.transaction_receipt['from'],
            "credit_to": transaction.transaction_receipt['to'],

        }
        return caaj_common

    @classmethod
    def __get_caaj_farms_unstake(cls, transaction):
        caaj_common = PancakePlugin.__get_caaj_common(transaction)

        debit_log = \
            list(filter(lambda log: log['topics'][0].hex().lower() == ERC20_TRANSFER_TOPIC
                        and "0x" + log['topics'][2].hex().lower()[26:] ==
                        transaction.transaction_receipt['from'].lower()
                        and "0x" + log['topics'][1].hex().lower()[26:] ==
                        transaction.transaction_receipt['to'].lower(),
                        transaction.transaction_receipt['logs']))[0]
        caaj_main = {
            "debit_title": "SPOT",
            "debit_amount": {debit_log['address']:
                             str(Decimal(int(debit_log['data'], 16))
                                 / Decimal(WEI))
                             },
            "credit_title": "STAKINGREWARD",
            "credit_amount": {debit_log['address']:
                              str(Decimal(int(debit_log['data'], 16))
                                  / Decimal(WEI))
                              },
            "comment": "pancakeswap reward"
        }
        caaj_main.update(caaj_common)
        return caaj_main

    @classmethod
    def __get_caaj_farms_reward(cls, transaction):
        debit_log_reward = \
            list(filter(lambda log: log['topics'][0].hex().lower() == ERC20_TRANSFER_TOPIC
                        and "0x" + log['topics'][2].hex().lower()[26:] ==
                        transaction.transaction_receipt['from'].lower()
                        and "0x" + log['topics'][1].hex().lower()[26:] ==
                        SYRUP_CONTRACT_ADDRESS.lower(),
                        transaction.transaction_receipt['logs']))[0]

        caaj_reward = {
            "time": transaction.get_timestamp(),
            "platform": cls.PLATFORM,
            "transaction_id": transaction.transaction_receipt['transactionHash'],
            "debit_title": "SPOT",
            "debit_from": SYRUP_CONTRACT_ADDRESS,
            "debit_to": transaction.transaction_receipt['from'],
            "debit_amount": {debit_log_reward['address']:
                             str(Decimal(int(debit_log_reward['data'], 16))
                                 / Decimal(WEI))
                             },
            "credit_title": "STAKINGREWARD",
            "credit_amount": {debit_log_reward['address']:
                              str(Decimal(int(debit_log_reward['data'], 16))
                                  / Decimal(WEI))
                              },
            "credit_from":  transaction.transaction_receipt['from'],
            "credit_to": SYRUP_CONTRACT_ADDRESS,
            "comment": "pancakeswap reward"
        }
        return caaj_reward

    @classmethod
    def __get_caaj_farms_stake(cls, transaction):
        caaj_common = PancakePlugin.__get_caaj_common(transaction)

        debit_log = \
            list(filter(lambda log: log['topics'][0].hex().lower() == ERC20_TRANSFER_TOPIC
                        and "0x" + log['topics'][1].hex().lower()[26:] ==
                        transaction.transaction_receipt['from'].lower()
                        and "0x" + log['topics'][2].hex().lower()[26:] ==
                        transaction.transaction_receipt['to'].lower(),
                        transaction.transaction_receipt['logs']))
        if len(debit_log) == 0:
            # harvest
            caaj_harvest = PancakePlugin.__get_caaj_farms_reward(transaction)
            return caaj_harvest
        else:
            # stake
            caaj_main = {
                "debit_title": "SPOT",
                "debit_amount": {debit_log[0]['address']:
                                 str(Decimal(int(debit_log[0]['data'], 16))
                                 / Decimal(WEI))
                                 },
                "credit_title": "STAKING",
                "credit_amount": {debit_log[0]['address']:
                                  str(Decimal(int(debit_log[0]['data'], 16))
                                  / Decimal(WEI))
                                  },
                "comment": "pancakeswap stake"
            }
            caaj_main.update(caaj_common)
            return caaj_main



from senka.senka import Senka
import os
import sys


if __name__ == '__main__':
  args = sys.argv
  chain = args[1]
  address = args[2]  
  setting = {}
  if chain == 'bsc':
    setting['bscscan_key'] = os.environ['BSCSCAN_KEY']
  senka = Senka(setting)
  caaj = senka.get_caaj_csv(chain, address)
  print(caaj)
  


from typing import List
from senkalib.senka_lib import SenkaLib
from senkalib.senka_setting import SenkaSetting
from senkalib.caaj_journal import CaajJournal
from senka.plugin_manager import PluginManager
import pandas as pd
import os

pd.set_option("display.max_columns", 50)

class Senka:
  def __init__(self, setting_dict):
    self.setting = SenkaSetting(setting_dict)
    pass
 
  def get_caaj_csv(self, chain:str, address:str) -> str:
    caaj_list = self.get_caaj(chain, address)
    caaj_dict_list = SenkaLib.get_caaj_jounal_dicts(caaj_list)
    df = pd.DataFrame(caaj_dict_list)
    df = df.sort_values('time')
    caaj_csv = df.to_csv(None, index=False)
    return caaj_csv

  def get_caaj(self, chain:str, address:str) -> List[CaajJournal]:
    token_original_ids = SenkaLib.get_token_original_ids()
    available_chains = self.get_available_chains()
    if chain.lower() not in available_chains:
      raise ValueError('this chain is not supported.')

    caaj = []
    transaction_generator = list(filter(lambda x:x.chain.lower() == chain.lower(),SenkaLib.get_available_chain()))[0]
    transactions = transaction_generator.get_transactions(self.setting, address)
    plugins = PluginManager.get_plugins(chain, self.get_plugin_dir_path())

    for transaction in transactions:
      for plugin in plugins:
        if plugin.can_handle(transaction):
          caaj_peace = plugin.get_caajs(address, transaction, token_original_ids)
          caaj.extend(caaj_peace)

    return caaj



  def get_available_chains(self) -> List[str]:
    chains = SenkaLib.get_available_chain()
    chains = list(map(lambda x:x.chain,chains))
    return chains

  def get_plugin_dir_path(self) -> str:
    return '%s/../plugin' % os.path.dirname(__file__)


import logging

class TestConfig:
  CRITICAL = logging.CRITICAL
  ERROR = logging.ERROR
  WARNING = logging.WARNING
  INFO = logging.INFO
  DEBUG = logging.DEBUG
  NOTEST = logging.NOTSET

  @classmethod
  def set_root_logger(cls, level):
    root_logger = logging.getLogger(name=None)
    root_logger.setLevel(level)
    if not root_logger.hasHandlers():
      fmt = logging.Formatter("%(asctime)s, %(name)s, %(lineno)s, %(levelname)s, %(message)s", "%Y-%m-%dT%H:%M:%S")
      stream_handler = logging.StreamHandler()
      stream_handler.setLevel(level)
      stream_handler.setFormatter(fmt)
      root_logger.addHandler(stream_handler)


import unittest
from unittest.mock import *
from senka.plugin_manager import PluginManager
from pathlib import Path
import os

print('%s/plugin' % os.path.dirname(__file__))
class TestPluginManager(unittest.TestCase):
  def test_get_plugins(self):
    plugins = PluginManager.get_plugins('test_one', '%s/plugin' % os.path.dirname(__file__))
    assert len(plugins) == 1
    assert plugins[0].chain == 'test_one'

  def test_get_plugins_exception(self):
    with self.assertRaises(RuntimeError):
      plugins = PluginManager.get_plugins('test_none', '%s/no_plugin' % os.path.dirname(__file__))

if __name__ == '__main__':
  unittest.main()


from setuptools import setup, find_packages
from glob import glob
from os.path import splitext
from os.path import basename

def _requires_from_file(filename):
    return open(filename).read().splitlines()

setup(
    name='senka',
    version='0.0.4',
    license='mit',
    description='making journal for transactions on blockchain',

    author='ca3-caaip',
    author_email='',
    url='https://github.com/ca3-caaip/senka',
    install_requires=_requires_from_file('requirements.txt'),
    extras_require={
        "test": ["pytest", "pytest-cov"]
    },
    packages=find_packages('src'),
    package_dir={'': 'src'},
    py_modules=[splitext(basename(path))[0] for path in glob('src/*.py')]
)



import unittest
from unittest.mock import *
from senka.senka import Senka
from senkalib.senka_lib import SenkaLib
from senkalib.chain.transaction_generator import TransactionGenerator
from senkalib.senka_setting import SenkaSetting
from senkalib.chain.transaction import Transaction
from senkalib.caaj_journal_amount import CaajJournalAmount
from typing import List
from decimal import Decimal
import os
import json
from pathlib import Path
import pandas as pd

class TestSenka(unittest.TestCase):
  def test_get_available_chains(self):
    with patch.object(SenkaLib, 'get_available_chain', new=TestSenka.mock_get_available_chains):
      senka = Senka({})
      chains = senka.get_available_chains()
      assert chains[0] == 'test_one'
      assert chains[1] == 'test_zero'

  def test_get_caaj_no_chain(self):
    with patch.object(SenkaLib, 'get_available_chain', new=TestSenka.mock_get_available_chains):
      with self.assertRaises(ValueError):
        senka = Senka({})
        caaj = senka.get_caaj('no_existence_chain', '0xfFceBED170CE0230D513a0a388011eF9c2F97503')

  def test_get_caaj(self):
    with patch.object(SenkaLib, 'get_available_chain', new=TestSenka.mock_get_available_chains):
      with patch.object(Senka, 'get_plugin_dir_path', new=TestSenka.mock_get_plugin_dir_path):
        with patch.object(SenkaLib, 'get_token_original_ids', new=TestSenka.mock_get_token_original_ids):
          senka = Senka({})
          caaj_list = senka.get_caaj('test_one', '0xfFceBED170CE0230D513a0a388011eF9c2F97503')
          cj = caaj_list[0]

          amount = CaajJournalAmount('testone', 'ibc/46B44899322F3CD854D2D46DEEF881958467CDD4B3B10086DA49296BBED94BED', 
            '3a2570c5-15c4-2860-52a8-bff14f27a236', Decimal('0.005147'))

          assert len(caaj_list) == 2
          assert cj.meta.time == "2022-01-01 00:00:00"
          assert cj.meta.platform == "test_platform"
          assert cj.meta.transaction_id == "0x1111111111111111111111111111"
          assert cj.meta.comment == "no comment"
          assert cj.debit.side_from == "0x0000000000000"
          assert cj.debit.side_to == "0x11111111111"
          assert cj.credit.side_from == "0x22222222222"
          assert cj.credit.side_to == "0x33333333333"
          assert cj.debit.title == "SPOT"
          assert cj.debit.amounts[0].symbol == 'testone'
          assert cj.debit.amounts[0].original_id == 'ibc/46B44899322F3CD854D2D46DEEF881958467CDD4B3B10086DA49296BBED94BED'
          assert cj.debit.amounts[0].symbol_uuid == '3a2570c5-15c4-2860-52a8-bff14f27a236'
          assert cj.credit.title == "SPOT"
          assert cj.credit.amounts[0].symbol == 'testone'
          assert cj.credit.amounts[0].original_id == 'ibc/46B44899322F3CD854D2D46DEEF881958467CDD4B3B10086DA49296BBED94BED'
          assert cj.credit.amounts[0].symbol_uuid == '3a2570c5-15c4-2860-52a8-bff14f27a236'

  def test_get_caaj_csv(self):
    with patch.object(SenkaLib, 'get_available_chain', new=TestSenka.mock_get_available_chains):
      with patch.object(Senka, 'get_plugin_dir_path', new=TestSenka.mock_get_plugin_dir_path):
        with patch.object(SenkaLib, 'get_token_original_ids', new=TestSenka.mock_get_token_original_ids):
          senka = Senka({})
          caaj_csv = senka.get_caaj_csv('test_one', '0xfFceBED170CE0230D513a0a388011eF9c2F97503')
          caaj_csv_lines = caaj_csv.splitlines()
          assert len(caaj_csv_lines) == 3
          assert caaj_csv_lines[0] == 'time,platform,transaction_id,debit_title,debit_amounts,debit_from,debit_to,\
credit_title,credit_amounts,credit_from,credit_to,comment'
          assert caaj_csv_lines[1] == '2022-01-01 00:00:00,test_platform,0x1111111111111111111111111111,\
SPOT,\
"[{\'token\': \
{\'symbol\': \'testone\', \
\'original_id\': \
\'ibc/46B44899322F3CD854D2D46DEEF881958467CDD4B3B10086DA49296BBED94BED\', \
\'symbol_uuid\': \'3a2570c5-15c4-2860-52a8-bff14f27a236\'}, \
\'amount\': Decimal(\'0.005147\')}]",0x0000000000000,0x11111111111,\
SPOT,\
"[{\'token\': \
{\'symbol\': \'testone\', \
\'original_id\': \
\'ibc/46B44899322F3CD854D2D46DEEF881958467CDD4B3B10086DA49296BBED94BED\', \
\'symbol_uuid\': \'3a2570c5-15c4-2860-52a8-bff14f27a236\'}, \
\'amount\': Decimal(\'0.005147\')}]",0x22222222222,0x33333333333,no comment'
          assert caaj_csv_lines[2] == '2022-01-01 00:00:00,test_platform,0x1111111111111111111111111111,\
SPOT,\
"[{\'token\': \
{\'symbol\': \'testone\', \
\'original_id\': \
\'ibc/46B44899322F3CD854D2D46DEEF881958467CDD4B3B10086DA49296BBED94BED\', \
\'symbol_uuid\': \'3a2570c5-15c4-2860-52a8-bff14f27a236\'}, \
\'amount\': Decimal(\'0.005147\')}]",0x0000000000000,0x11111111111,\
SPOT,\
"[{\'token\': \
{\'symbol\': \'testone\', \
\'original_id\': \
\'ibc/46B44899322F3CD854D2D46DEEF881958467CDD4B3B10086DA49296BBED94BED\', \
\'symbol_uuid\': \'3a2570c5-15c4-2860-52a8-bff14f27a236\'}, \
\'amount\': Decimal(\'0.005147\')}]",0x22222222222,0x33333333333,no comment'

  @classmethod
  def mock_get_available_chains(cls):
    return [OneTransactionGenerator, ZeroTransactionGenerator]

  @classmethod
  def mock_get_token_original_ids(cls):
    df = pd.read_csv("test/testdata/token_original_ids/token_original_id.csv")
    return df

  @classmethod
  def mock_get_plugin_dir_path(cls):
    return '%s/plugin' % os.path.dirname(__file__)

class OneTransactionGenerator(TransactionGenerator):
  chain = 'test_one'
  @classmethod
  def get_transactions(cls, settings:SenkaSetting, address:str, timerange:dict = None, blockrange:dict = None) -> List[Transaction]:
    header = json.loads(Path('%s/testdata/header.json' % os.path.dirname(__file__)).read_text())
    receipt = json.loads(Path('%s/testdata/approve.json' % os.path.dirname(__file__)).read_text())
    transaction_a = OneTransaction(header['hash'], receipt, header['timeStamp'], header['gasUsed'], header['gasPrice'])
    transaction_b = OneTransaction(header['hash'], receipt, header['timeStamp'], header['gasUsed'], header['gasPrice'])
    return [transaction_a, transaction_b]

class OneTransaction(Transaction):
  chain = 'test_one'

  def __init__(self, transaction_id:str, transaction_receipt:dict, timestamp:str, gasused:str, gasprice:str):
    super().__init__(transaction_id)
    self.transaction_receipt = transaction_receipt
    self.timestamp = timestamp
    self.gasused = Decimal(gasused)
    self.gasprice = Decimal(gasprice)
  def get_timestamp(self) -> str:
    pass

  def get_transaction_fee(self) -> Decimal:
    pass

class ZeroTransactionGenerator(TransactionGenerator):
  chain = 'test_zero'
  @classmethod
  def get_transactions(cls, settings:SenkaSetting, address:str, timerange:dict = None, blockrange:dict = None) -> List[Transaction]:
    pass

if __name__ == '__main__':
  unittest.main()